﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion


using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  [Item("CrossoverPerformanceAnalyzer", "An operator that analyzes the performance of crossovers.")]
  [StorableClass]
  public class CrossoverPerformanceAnalyzer : InitializableOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string GenerationsParameterName = "Generations";
    private const string SuccessfullCrossoversRowName = "Successfull Crossovers per Generation with CompFact ";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion

    #region Parameter properties
    public IValueLookupParameter<BoolValue> MaximizationParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public IValueParameter<ItemCollection<DoubleValue>> ComparisonFactorParameter {
      get { return (IValueParameter<ItemCollection<DoubleValue>>)Parameters["ComparisonFactor"]; }
    }
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    public ILookupParameter<ItemArray<Permutation>> ParentsParameter {
      get { return (ScopeTreeLookupParameter<Permutation>)Parameters["Parents"]; }
    }
    public ILookupParameter<ItemArray<DoubleValue>> ParentsQualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["ParentsQuality"]; }
    }
    public ILookupParameter<Permutation> ChildParameter {
      get { return ((LookupParameter<Permutation>)Parameters["Child"]); }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public IValueParameter<ISingleObjectiveSolutionSimilarityCalculator> SimilarityCalculatorParameter {
      get { return (IValueParameter<ISingleObjectiveSolutionSimilarityCalculator>)Parameters["SimilarityCalculator"]; }
    }
    public ILookupParameter<ItemCollection<IItem>> OperatorsParameter {
      get { return (ILookupParameter<ItemCollection<IItem>>)Parameters["Operators"]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    #endregion

    [Storable]
    private ScatterPlotHelper plotHelper, childDiversityHelper, parentDiversityHelper;
    [Storable]
    private DataTableHelper performanceHelper, successHelper, equalParentsHelper;
    [Storable]
    private int cnt = 0;
    [Storable]
    private int[] success;
    [Storable]
    private int lastGeneration = 0;
    [Storable]
    private int equalParents = 0;
    [Storable]
    private List<double> qualityPoints = new List<double>();

    [StorableConstructor]
    private CrossoverPerformanceAnalyzer(bool deserializing) : base(deserializing) { }
    private CrossoverPerformanceAnalyzer(CrossoverPerformanceAnalyzer original, Cloner cloner)
      : base(original, cloner) {
      cnt = original.cnt;
      success = (int[])original.success.Clone();
      lastGeneration = original.lastGeneration;
      equalParents = original.equalParents;
      qualityPoints = new List<double>(original.qualityPoints);
      plotHelper = (ScatterPlotHelper)original.plotHelper.Clone(cloner);
      childDiversityHelper = (ScatterPlotHelper)original.childDiversityHelper.Clone(cloner);
      parentDiversityHelper = (ScatterPlotHelper)original.parentDiversityHelper.Clone(cloner);
      performanceHelper = (DataTableHelper)original.performanceHelper.Clone(cloner);
      successHelper = (DataTableHelper)original.successHelper.Clone(cloner);
      equalParentsHelper = (DataTableHelper)original.equalParentsHelper.Clone(cloner);
    }

    public CrossoverPerformanceAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "Nr of generations."));

      Parameters.Add(new ScopeTreeLookupParameter<Permutation>("Parents", "The parent permutations which have been crossed."));
      ParentsParameter.ActualName = "TSPTour";

      Parameters.Add(new LookupParameter<Permutation>("Child", "The child permutation resulting from the crossover."));
      ChildParameter.ActualName = "TSPTour";

      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("ParentsQuality", "The quality of the parent solutions."));
      ParentsQualityParameter.ActualName = "TSPTourLength";

      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The evaluated quality of the child solution."));
      QualityParameter.ActualName = "TSPTourLength";

      Parameters.Add(new ValueParameter<ISingleObjectiveSolutionSimilarityCalculator>("SimilarityCalculator"));
      Parameters.Add(new LookupParameter<ItemCollection<IItem>>("Operators", "The operators and items that the problem provides to the algorithms."));
      Parameters.Add(new ValueLookupParameter<BoolValue>("Maximization", "True if the problem is a maximization problem, false otherwise"));
      Parameters.Add(new ValueParameter<ItemCollection<DoubleValue>>("ComparisonFactor", "Determines if the quality should be compared to the better parent (1.0), to the worse (0.0) or to any linearly interpolated value between them."));
      ComparisonFactorParameter.Value.Add(new DoubleValue(0));
      ComparisonFactorParameter.Value.Add(new DoubleValue(0.5));
      ComparisonFactorParameter.Value.Add(new DoubleValue(1));


      plotHelper = new ScatterPlotHelper(false, true);
      childDiversityHelper = new ScatterPlotHelper(false, true);
      parentDiversityHelper = new ScatterPlotHelper(false, true);
      performanceHelper = new DataTableHelper();
      successHelper = new DataTableHelper();
      equalParentsHelper = new DataTableHelper();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CrossoverPerformanceAnalyzer(this, cloner);
    }

    protected override void InitializeAction() {
      if (SimilarityCalculatorParameter.Value == null) {
        SimilarityCalculatorParameter.Value = OperatorsParameter.ActualValue.OfType<ISingleObjectiveSolutionSimilarityCalculator>().FirstOrDefault();
      }

      if (success == null) {
        success = new int[ComparisonFactorParameter.Value.Count];
      }

      plotHelper.InitializePlot(Results, "Crossover Performance", "Solution Index", "Absolut Quality Difference");
      childDiversityHelper.InitializePlot(Results, "Child Diversity", "Solution Index", "Diversity");
      parentDiversityHelper.InitializePlot(Results, "Parent Diversity", "Solution Index", "Diversity");
      performanceHelper.InitializeChart(Results, "Average Crossover Performance", new string[] { "Average Crossover Performance per Generation" });
      equalParentsHelper.InitializeChart(Results, "Number of equal parents", new string[] { "Absolut number of equal parents" });

      List<string> successfullCXRowNames = new List<string>();
      foreach (var value in ComparisonFactorParameter.Value) {
        successfullCXRowNames.Add(SuccessfullCrossoversRowName + value.Value.ToString());
      }
      successHelper.InitializeChart(Results, "Successfull Crossovers", successfullCXRowNames.ToArray());
    }

    public override IOperation Apply() {
      Initialize();

      Point2D<double> qualityPoint, diversityPointChild, diversityPointParent;
      var qualityParent1 = ParentsQualityParameter.ActualValue.First().Value;
      var qualityParent2 = ParentsQualityParameter.ActualValue.Last().Value;
      var qualityChild = QualityParameter.ActualValue.Value;
      var child = ChildParameter.ActualValue;
      var parent1 = ParentsParameter.ActualValue.First();
      var parent2 = ParentsParameter.ActualValue.Last();
      var parentDiversity = SimilarityCalculatorParameter.Value.CalculateSolutionSimilarity(ExecutionContext.Scope.SubScopes.First(), ExecutionContext.Scope.SubScopes.Last());
      string curGenStr = GenerationsParameter.ActualValue.Value.ToString();

      diversityPointParent = new Point2D<double>(cnt, parentDiversity);
      double worseQuality = qualityParent1 > qualityParent2 ? qualityParent1 : qualityParent2;
      if (qualityParent1 > qualityParent2) {
        diversityPointChild = new Point2D<double>(cnt, SimilarityCalculatorParameter.Value.CalculateSolutionSimilarity(ExecutionContext.Scope, ExecutionContext.Scope.SubScopes.First()));
      } else {
        diversityPointChild = new Point2D<double>(cnt, SimilarityCalculatorParameter.Value.CalculateSolutionSimilarity(ExecutionContext.Scope, ExecutionContext.Scope.SubScopes.Last()));
      }

      qualityPoint = new Point2D<double>(cnt++, worseQuality - qualityChild);
      qualityPoints.Add(qualityPoint.Y);

      //track successfull cx
      var compFactsArray = ComparisonFactorParameter.Value.ToArray();
      for (int i = 0; i < compFactsArray.Length; i++) {
        if (WeightedParentsQualityComparer.Compare(qualityChild, qualityParent1, qualityParent2, compFactsArray[i].Value, MaximizationParameter.ActualValue.Value)) {
          success[i]++;
        }
      }

      if (parentDiversity == 1.0) {
        equalParents++;
      }

      if (GenerationsParameter.ActualValue.Value != 0) {
        if (GenerationsParameter.ActualValue.Value > lastGeneration) {
          double avg = qualityPoints.Average();

          performanceHelper.AddPoint(avg);
          equalParentsHelper.AddPoint(equalParents);

          for (int i = 0; i < compFactsArray.Length; i++) {
            successHelper.AddPoint(SuccessfullCrossoversRowName + compFactsArray[i].Value.ToString(), success[i]);
          }

          Reset();
        }

        plotHelper.AddPoint(curGenStr, qualityPoint);
        childDiversityHelper.AddPoint(curGenStr, diversityPointChild);
        parentDiversityHelper.AddPoint(curGenStr, diversityPointParent);
      } else {
        Reset();
      }

      return base.Apply();
    }

    private void Reset() {
      cnt = 0;
      success = new int[ComparisonFactorParameter.Value.Count];
      lastGeneration = GenerationsParameter.ActualValue.Value;
      qualityPoints.Clear();
      equalParents = 0;
    }
  }
}
